#
# Copyright 2011, Pall Sigurdsson <palli@opensource.is>
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module provides helper functions for okconfig """

from __future__ import absolute_import
from __future__ import print_function
from builtins import object
import re
import fcntl
import os
from subprocess import Popen, PIPE, STDOUT

from pynag import Model

import okconfig


def add_defaultservice_to_host(host_name):
    """ Given a specific hostname, add default service to it """
    # Get our host
    try: my_host = Model.Host.objects.get_by_shortname(host_name)
    except ValueError: raise okconfig.OKConfigError("Host %s not found." % host_name)

    # Dont do anything if file already exists
    service = Model.Service.objects.filter(name=host_name)
    if len(service) != 0:
        return False

    # Make sure that host belongs to a okconfig-compatible group
    hostgroup_name = my_host['hostgroups'] or "default"
    hostgroup_name = hostgroup_name.strip('+')
    if hostgroup_name in okconfig.get_groups():
        GROUP=hostgroup_name
    else:
        GROUP='default'

    template = default_service_template
    template = re.sub("HOSTNAME", host_name, template)
    template = re.sub("GROUP", GROUP, template)
    fh = open( my_host['filename'], 'a')
    fh.write(template)
    fh.close()
    return True


def group_exists(group_name):
    """ Check if a servicegroup,contactgroup or hostgroups exist with shortname == group_name

    Returns:
        False if no groups are found, otherwise it returns a list of groups
    """
    servicegroups = Model.Servicegroup.objects.filter(shortname=group_name)
    contactgroups = Model.Contactgroup.objects.filter(shortname=group_name)
    hostgroups = Model.Hostgroup.objects.filter(shortname=group_name)
    result = (servicegroups+contactgroups+hostgroups)
    if result == ([]):
        return False
    return result

def runCommand(command):
    """runCommand: Runs command from the shell prompt.

     Arguments:
         command: string containing the command line to run
     Returns:
         stdout/stderr of the command run
     Raises:
         BaseException if returncode > 0
    """
    proc = Popen(command, shell=True, stdin=PIPE,stdout=PIPE,stderr=PIPE,)
    stdout, stderr = proc.communicate('through stdin to stdout')
    result = proc.returncode,stdout,stderr
    if proc.returncode > 0:
        error_string = "* Could not run command (return code= %s)\n" % proc.returncode
        error_string += "* Error was:\n%s\n" % (stderr.strip())
        error_string += "* Command was:\n%s\n" % command
        error_string += "* Output was:\n%s\n" % (stdout.strip())
        if proc.returncode == 127: # File not found, lets print path
            path=os.getenv("PATH")
            error_string += "Check if y/our path is correct: %s" % path
        raise okconfig.OKConfigError( error_string )
    else:
        return result


class clientInstall(object):
    import okconfig.config
    def __init__(self,
                 script=okconfig.config.nsclient_installfiles + "/install_nsclient.sh",
                 script_args=None,
                 merge_env=None):
        """
        Initializes the object for nsclient installs

        :param host_name: hostname of machine to install on
        :param domain:    windows domain, use '.' for local domain
        :param username:  username with privileges to install
        :param password:  yeah, the password
        """
        self.process = None
        self.unparseable_lines = []
        self.stage_state = {}
        self.script = script
        self.script_args = script_args
        self.current_stage = None
        self.merge_env = merge_env

    def execute(self):
        """
        Executes the install_nsclient command which installs the nsclient agent for windows machines
        """

        user_env = os.environ
        if self.merge_env:
            user_env.update(self.merge_env)

        if not self.script_args:
            self.script_args = []

        try:
            self.process = Popen([self.script] + self.script_args,
                                 env=user_env,
                                 stdout=PIPE,
                                 stderr=STDOUT,
                                 bufsize=1,
                                 shell=False
            )

            # Make stdout non blocking
            fd = self.process.stdout.fileno()
            flags = fcntl.fcntl(fd, fcntl.F_GETFL)
            fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        except Exception as e:
            raise okconfig.OKConfigError(e)

    def get_state(self):
        """
        Checks if there is any output from the script and parses it
        return the state of each install stage
        """
        import re

        while True:
            self.process.stdout.flush()
            line = ""
            try:
                line = self.process.stdout.readline()
            except IOError:
                break
            if not line:
                break
            m = re.match("^\[(.*?)\s*\] (\S+?) (.*)$", line)
            if m:
                self.stage_state[m.group(1)] = m.group(3)
                self.current_stage = m.group(1)
            else:
                self.unparseable_lines.append(line)
        return self.current_stage, self.stage_state


default_service_template = '''
# This is a template service for HOSTNAME
# Services that belong to this host should use this as a template
define service {
        name                            HOSTNAME
        use                             GROUP-default_service
        host_name                       HOSTNAME
        contact_groups                  +GROUP
        service_groups                  +GROUP
        service_description             Default Service for HOSTNAME
        register                        0
}
'''

if __name__ == '__main__':
    print(add_defaultservice_to_host('host1'))



#
# Copyright 2011, Pall Sigurdsson <palli@opensource.is>
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides an interface to okconfig utilities
and operations such as adding new hosts to Nagios or adding
templates to current hosts

Example Usage:

import okconfig
okconfig.nagios_config="/etc/nagios/nagios.cfg"
okconfig.addhost("myhost.example.com", group_name="databases", templates=["linux","mysql"])
"""

from __future__ import absolute_import
__author__ = "Pall Sigurdsson"
__copyright__ = "Copyright 2011, Pall Sigurdsson"
__credits__ = ["Pall Sigurdsson"]
__license__ = "GPL"
__version__ = '1.4.2'
__maintainer__ = "Gardar Thorsteinsson"
__email__ = "gardar@ok.is"
__status__ = "Development"


import socket
import os
try:
    import paramiko
except ImportError:
    paramiko = None

import pynag.Model
import pynag.Utils

from . import network_scan
from . import config
from . import helper_functions

nagios_config = config.nagios_config
template_directory =config.template_directory
examples_directory= config.examples_directory
examples_directory_local= config.examples_directory_local
destination_directory = config.destination_directory
pynag.Model.pynag_directory = destination_directory

required_gateways = []

pynag.Model.cfg_file = config.nagios_config

version = __version__
def is_valid():
    """Checks if okconfig is installed and properly configured.

    Returns True/False

    See verify() for more details
    """
    checks = verify()
    for result in list(checks.values()):
        if result is False: return False
    return True

def _is_in_path(command):
    """ Searches $PATH and returns true if command is found, and in path """


    is_executable = lambda x: os.path.isfile(x) and os.access(x, os.X_OK)

    if command.startswith('/'):
        return is_executable(command)

    my_path = os.environ['PATH'].split(':')
    for possible_path in my_path:
        full_path = "%s/%s" % (possible_path,command)
        if is_executable(full_path): return True
    return False


def verify():
    """Checks if okconfig is installed and properly configured.

    Returns dict of {'check_name':Boolean}

    Check if:
    1) nagios_config exists
    2) template_directory exists
    3) destination_directory exists (and is writable)
    """
    results = {}

    # 1) nagios_config exists
    check = "Main configuration file %s is readable" % nagios_config
    results[check] = os.access(nagios_config, os.R_OK)

    # 2) template_directory exists
    check = "template_directory %s exists" % template_directory
    results[check] = os.access(template_directory, os.R_OK) and os.path.isdir(template_directory)

    # 3) destination_directory or parent exists (and is writable)
    for ddir in [destination_directory, "%s/.." % destination_directory]:
        ddir = os.path.dirname(ddir)
        check = "destination_directory %s is writable" % ddir
        results[check] = os.access(ddir, os.W_OK + os.R_OK) and os.path.isdir(ddir)
        if results[check] == True: break

    # 4)
    # Should no longer be need
    # TODO: Remove this commented codeblock
    #okconfig_binaries = ('addhost','findhost','addgroup','addtemplate')
    #for command in okconfig_binaries:
    #	check = "'%s' command is in path" % command
    #	results[check] = _is_in_path(command)

    return results

def addhost(host_name, address=None, group_name=None, templates=None, use=None, alias=None, host_template='host', force=False):
    """Adds a new host to Nagios. Returns true if operation is successful.

    Args:
     host_name -- Hostname of the host to be added
     address -- IP Address of the host (if None, it will be looked up in DNS)
     group_name -- Primary host/contactgroup for this host. (if none, use "default")
     templates -- List of template names to be added to this host
     use -- if this host inherits another host (i.e. "windows-server")
     host_template -- Use the specified host template instead of default host.cfg
     force -- Force operation. Overwrite config files needed.

    Examples:
    >>> addhost(host_name="localhost",group_name="database_servers") # doctest: +SKIP
    >>> addhost(host_name="okconfig.org", address="192.168.1.1") # doctest: +SKIP

    Returns:
     String message with result of addhost
    """
    if group_name is None or group_name is '': group_name = 'default'
    if templates is None: templates = []
    if alias is None: alias = host_name
    if address is None or address is '':
        try:
            address = socket.gethostbyname(host_name)
        except:
            raise OKConfigError("Could not resolve hostname '%s'" % host_name)
    if use is None:
        use = 'okc-default-host'
    okconfig_groups = get_groups()
    if len(okconfig_groups) == 0:
        addgroup(group_name='default',alias='OKconfig default group')
    arguments = {'PARENTHOST': use, 'GROUP': group_name, 'IPADDR': address, 'HOSTNAME': host_name, 'ALIAS': alias}
    destination_dir = "%s/hosts/%s/" % (destination_directory, group_name)
    destination_file = "%s/%s-host.cfg" % (destination_dir, host_name)
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    if not force:
        if os.path.isfile(destination_file):
            raise OKConfigError("Destination file '%s' already exists." % destination_file)
        if group_name not in get_groups():
            #raise OKConfigError("Group %s does not exist" % group_name)
            addgroup(group_name)
        if host_name in get_hosts():
            filename = pynag.Model.Host.objects.get_by_shortname(host_name)._meta['filename']
            raise OKConfigError("Host named '%s' already exists in %s" % (host_name, filename))
    # Do sanity checking of all templates before we add anything
    all_templates = list(get_templates().keys())
    if host_template not in all_templates:
        raise OKConfigError("Host Template %s not found" % host_template)
    for i in templates:
        if i not in all_templates:
            raise OKConfigError("Template %s not found" % i)
    result = _apply_template(host_template, destination_file, **arguments)
    _git_commit(filelist=result, message='okconfig host %s added with address %s' % (host_name,address))
    for i in templates:
        result = result + addtemplate(host_name=host_name, template_name=i, group_name=group_name,force=force)
    return result
def addtemplate(host_name, template_name, group_name=None,force=False):
    """Adds a new template to existing host in Nagios.

    Args:
     host_name -- Hostname to add template to (i.e. "host.example.com")
     template_name -- Name of the template to be added (i.e. "mysql")
     force -- Force operation, overwrites configuration if it already exists

    Examples:
     addtemplate(host_name="host.example.com", template="mysql")

    Returns:
     True if operation is succesful.
    """
    try:
        host = pynag.Model.Host.objects.get_by_shortname(host_name)
    except (ValueError, KeyError):
        raise OKConfigError("Host '%s' was not found" % host_name)
    hostfile = host.get_filename()

    # If no group name is specified, try host.contact_groups
    if group_name is None: group_name = host.contact_groups
    if group_name is None: group_name = "default"

    if hostfile is None:
        raise OKConfigError("Host '%s' was not found" % host_name)
    if template_name not in list(get_templates().keys()):
        raise OKConfigError("Template '%s' was not found" % template_name)
    hostdir = os.path.dirname(hostfile)
    # Check if host has the required "default service"
    helper_functions.add_defaultservice_to_host(host_name)

    # Lets do some templating
    newfile = "%s/%s-%s.cfg" % (hostdir, host_name,template_name)
    if not force:
        # 'Do some basic sanity checking'
        if os.path.exists(newfile):
            raise OKConfigError("Destination file '%s' already exists." % newfile)

    result = _apply_template(template_name,newfile, HOSTNAME=host_name, GROUP=group_name)
    _git_commit(filelist=result, message='okconfig template %s added to host %s' % (template_name, host_name))
    return result

def addgroup(group_name, alias=None, force=False):
    """Adds a new hostgroup/contactgroup/servicegroup combo to Nagios.

    Args:
     group_name -- Name of the group to be added (i.e. "db-servers")
     alias -- Group alias (i.e. "Database Servers")
     force -- Force operation, overwrites configuration if it already exists

    Examples:
     addgroup(group_name="db-servers", alias="Database Servers")

    Returns:
     True if operation was successful
    """
    if alias is None: alias=group_name
    destination_dir = "%s/groups" % destination_directory
    destination_file = "%s/%s.cfg" % (destination_dir, group_name)
    if not force:
        # 'Do some sanity checking'
        if os.path.exists(destination_file):
            raise OKConfigError("Destination file '%s' already exists" % destination_file)
        groups = helper_functions.group_exists(group_name)
        if groups != False:
            raise OKConfigError("We already have groups with name = %s" % group_name)

    result = _apply_template(template_name="group",destination_file=destination_file, GROUP=group_name, ALIAS=alias)
    _git_commit(filelist=result, message="okconfig group %s added" % group_name)
    return result
def addcontact(contact_name, alias=None, force=False, group_name="default", email=None, use='generic-contact'):
    """Adds a new contact to Nagios.

    Args:
     contact_name -- Name of the contact to be added (i.e. "user@example.com")
     alias -- Contact alias (i.e. "Full Name")
     force -- Force operation, overwrites configuration if it already exists

    Examples:
     addcontact(contact_name="user@example.com", contact_groups="default", email="user@example.com")

    Returns:
     True if operation was successful
    """
    if group_name is None or group_name is '': group_name = 'default'
    # Check if contact already exists
    try:
        contact = pynag.Model.Contact.objects.get_by_shortname(contact_name)
        if not force:
            raise OKConfigError("contact %s already exists in file %s" % (contact_name, contact.get_filename()))
    except (KeyError, ValueError):
        contact = pynag.Model.Contact()
    contact['contact_name'] = contact_name
    if alias is not None: contact['alias'] = alias
    if use is not None: contact['use'] = use
    if email is not None: contact['email'] = email
    if group_name is not None: contact['contactgroups'] = group_name
    result = contact.save()
    filename = contact.get_filename()
    if result is False:
        raise OKConfigError("Failed to save contact to %s" % filename)
    return [filename]
def addservice(inherit_settings_from, host_name, service_description=None, group=None, check_command=None,force=False):
    """Adds a new service to Nagios.

    Args:
     inherit_settings_from -- Name of another service to inherit from (i.e. 'generic-service')
     service_description -- Description of the service as it will appear in nagios (i.e. "Disk Usage")
     host_name -- Host that this service check belongs to
     force -- Force operation, overwrites configuration if it already exists

    Examples:
     addservice(host_name="localhost, inherit_settings_from="okc-check_ping")

    Returns:
     True if operation was successful
    """

    # Check if host exists:
    tmp = pynag.Model.Host.objects.filter(host_name=host_name)
    if len(tmp) != 1:
        raise OKConfigError("Cannot find host named '%s'." % host_name)
    host = tmp[0]

    # Find a proper filename to save to
    dirname = os.path.dirname( host.get_filename() )
    filename = '%s/%s-custom.cfg' % (dirname, host_name)

    # Check if parent exists
    tmp = pynag.Model.Service.objects.filter(name=inherit_settings_from)
    if len(tmp) != 1:
        raise OKConfigError("cannot find service '%s' to inherit settings from" % inherit_settings_from)
    parent_service = tmp[0]

    if service_description is None:
        if parent_service.service_description is None:
            raise OKConfigError("service_description not defined and parent '%s' does not have any." % inherit_settings_from)
        service_description = parent_service.service_description
    tmp = pynag.Model.Service.objects.filter(host_name=host_name,service_description=service_description)
    if len(tmp) is 0: # Service not found. This is indeed a new service
        service = pynag.Model.Service()
        service._meta['filename'] = filename
    elif len(tmp) is 1: # There is another service defined just like this one
        service = tmp[0]
        if not force:
            filename = service.get_filename()
            raise OKConfigError("service is already defined in file %s. Will not add again." % filename)
    else: # Multiple services defined
        files = []
        for i in tmp:
            files.append(i.get_filename())
        raise OKConfigError("%s services already defined with same host_name/service_description. Will not add any more: %s " % (len(tmp), "; ".join(files)))

    service['host_name'] = host_name
    if service_description is not None:
        service['service_description'] = service_description
    if group is not None:
        service['contact_groups'] = group
    if check_command is not None:
        service['check_command'] = check_command
    result = service.save()
    if result is False:
        raise OKConfigError("Failed to save service to file" % service.get_filename())
    else:
        return [service.get_filename()]

def findhost(host_name):
    """Returns the filename which defines a specied host. Returns None on failure.

    Args:
     host_name -- Name of the host to find

    Examples:
    >>> print findhost("host.example.com") # doctest: +SKIP
    "/etc/okconfig/hosts/default/host.example.com-host.cfg"
    """
    try:
        my_host = pynag.Model.Host.objects.get_by_shortname(host_name)
        filename = my_host['meta']['filename']
        return filename
    except (ValueError, KeyError):
        return None

def removehost(host_name, recursive=True):
    """ Removes a specified host. And possibly all services for that host.

    Args:
        host_name -- Name of the host to remove
        recursive -- If true: Also delete all services that belong to this host.
    Examples:
    >>> removehost('host.example.com', recursive=True) # doctest: +SKIP
    """
    my_host = pynag.Model.Host.objects.get_by_shortname(host_name)
    my_host.delete(recursive=recursive)
    return True

def get_templates():
    """ Returns a list of available templates """
    result = {}
    if not os.path.isdir(examples_directory):
        raise OKConfigError("Examples directory does not exist: %s" % examples_directory)
    filelist = os.listdir(examples_directory)
    if os.path.isdir(examples_directory_local):
        for i in os.listdir(examples_directory_local):
            if i not in filelist:
                filelist.append(i)
    for file in filelist:
        if os.path.isfile(examples_directory + "/" + file): filename = examples_directory + "/" + file
        if os.path.isfile(examples_directory_local + "/" + file): filename = examples_directory_local + "/" + file
        if file.endswith('.cfg-example'):
            template_name = file[:-12]
            template_parents = []
            template_friendly_name = ''
            result[template_name] = {'parents':template_parents, 'filename':filename, 'name':template_friendly_name}
    return result

def get_hosts():
    """ Returns a list of available hosts """
    result = []
    hosts = pynag.Model.Host.objects.all
    for host in hosts:
        if host.get_shortname() not in result and host.get_shortname() is not None:
            result.append(host.get_shortname())
    return result

def get_groups():
    """ Returns a list of available groups """
    result = []
    group_directory = "%s/groups" % destination_directory
    # If group_directory does not exist, we dont have any groups
    if not os.path.isdir(group_directory):
        return result
    filelist = os.listdir(group_directory)
    for file in filelist:
        if os.path.isfile(group_directory + "/" + file) and file.endswith('.cfg'):
            name = file[:-4]
            result.append(name)
    result.sort()
    return result

def install_nsclient(remote_host, domain, username, password):
    """ Logs into remote (windows) host and installs NSClient.

    Args:
     remote_host -- Hostname/IPAddress of remote host
     username -- Name of a user with administrative privileges on the remote host
     password -- Password to use
     domain -- Windows Domain

    Returns:
     True if operation was successful. Otherwise False
    """
    if not network_scan.check_tcp(remote_host, 445, timeout=5):
        raise OKConfigError('Cannot reach remote_host on port 445, aborting...')

    result = pynag.Utils.runCommand("%s/install_nsclient.sh '%s' --domain '%s' --user '%s' --password '%s'" % (config.nsclient_installfiles,remote_host,domain,username,password))
    return result

def check_agent(host_name):
    """ Checks a remote host if it has a valid okconfig client configuration

    Args:
        host_name -- hostname (or ip address of remote host)
    Returns:
        True/False, [ "List","of","messages" ]
    """
    raise NotImplementedError()

def install_okagent(remote_host,username,password=None, domain=None, install_method=None):
    ''' Installs an okagent to remote host using either winexe or ssh method

    Args:
     remote_host    -- Hostname/IPAddress of remote host
     username       -- Username to use
     password       -- Password to use. If None, try to use ssh keys
     install_method -- Use either "winexe" or "ssh". Leave empty for autodetect.
    Returns:
     exit_status,stdout,stderr
    '''
    if not install_method or install_method == '':
        if network_scan.check_tcp(remote_host, 22, timeout=5):
            install_method = 'ssh'
        elif network_scan.check_tcp(remote_host, 445, timeout=5):
            install_method = 'winexe'

    if install_method == 'ssh':
        return install_nrpe(remote_host=remote_host,username=username, password=password)
    elif install_method == 'winexe':
        return install_nsclient(remote_host=remote_host,username=username, password=password, domain=domain)
    raise OKConfigError("Cannot connect to %s on port 22 or 445. No install method available" % (remote_host))

def install_nrpe(remote_host, username, password=None):
    """ Logs into remote (unix) host and install nrpe-client.

    Args:
     remote_host -- Hostname/IPAddress of remote host
     username -- Username to use
     password -- Password to use. If None, try to use ssh keys

    Returns:
     True if operation was successful.
    """
    if not paramiko:
        raise OKConfigError('You need to install python module: paramiko')
    if not network_scan.check_tcp(remote_host, 22, timeout=5):
        raise OKConfigError('Cannot reach remote_host on port 22, aborting...')
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy( paramiko.AutoAddPolicy() )
    ssh.connect(remote_host, username=username, password=password)


    # Try a test command on the remote server to see if we are connected
    test_command = 'echo -e connection test'
    chan = ssh.get_transport().open_session()
    chan.exec_command( test_command )
    exit_status = chan.recv_exit_status()
    stdout = chan.recv(-1).strip()
    stderr = chan.recv_stderr(-1).strip()
    if exit_status != 0:
        ssh.close()
        raise OKConfigError('Exit code %s when trying to run test command\noutput: %s\nstderr: %s' % (chan.exit_status, stdout,stderr) )

    # Uploading install script to remote server
    sftp = ssh.open_sftp()
    sftp.put(config.install_nrpe_script, 'install_nrpe.sh')

    # Executing remote script
    # We need to do some hoola-hoops to get the exit code
    chan = ssh.get_transport().open_session()
    chan.exec_command('bash install_nrpe.sh')
    exit_status = chan.recv_exit_status()
    stdout = chan.recv(-1).strip()
    stderr = chan.recv_stderr(-1).strip()

    return exit_status,stdout,stderr

def _apply_template(template_name,destination_file, **kwargs):
    """ Applies okconfig template to filename, doing replacements from kwargs in the meantime

    Arguments:
        template_name - name of the template to use
        destination_file - full path to file to be written to
        kwargs key/value pair of string to search and replacement to make

    Example:
        _apply_template('host','/etc/nagios/okconfig/hosts/newhost.cfg', HOSTNAME='newhost',ADDRESS='0.0.0.0',GROUP='default')
    Returns:
        List of filenames that have been written to
    """
    all_examples = get_templates()
    if template_name not in all_examples:
        raise OKConfigError('Template %s cannot be found' % template_name)
    sourcefile = all_examples[template_name]['filename']

    # Clean // from destination file
    destination_file = destination_file.replace('//','/')

    if not os.path.isfile(sourcefile):
        raise OKConfigError('Template %s cannot be found' % template_name)

    dirname = os.path.dirname(destination_file)
    if not os.path.exists(dirname): os.makedirs(dirname)

    fd = open(sourcefile).read()
    for old_string,new_string in list(kwargs.items()):
        fd = fd.replace(old_string,new_string)
    open(destination_file,'w').write( fd )
    return [destination_file]

def _git_commit(filelist, message):
    """ If config.git_commit_changes is enabled, then commit "filelist" to the repository using message """
    if config.git_commit_changes != '1':
        return
    if 'git' not in globals():
        from pynag.Utils import GitRepo
        git = GitRepo(directory=os.path.dirname(config.nagios_config), auto_init=False, author_name="okconfig")
    else:
        git = globals()['git']
    git.commit(message=message, filelist=filelist)
class OKConfigError(Exception):
    pass

#all_templates = get_templates()
if __name__ == '__main__':
    pass



#
# Copyright 2011, Pall Sigurdsson <palli@opensource.is>
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
TEMPLATE_VERSION HISTORY

# Version 2.5 (2013-06-28)
    * okc-check_http has new service macro: __ON_REDIRECT
# Version 2.4 (2013-05-13)
    * Removed the following hardcoded groups from templates:
    * hostgroup cisco-hostgroup
    * servicegroup proliant-services
    * contactgroup proliant-contacts
    * servicegroup eva-services
    * contactgroup eva-contacts
    * servicegroup windows-services
    * servicegroup linux-services
    * servicegroup mssql-services
# Version 2.3 (2013-04-29)
    * okc-emc-check_portstate now enforces username and password
    * okc-check_http now has support for virtualhosts and ports
# Version 2.2 (2012-09-26)
    * Brocade templates reworked. _SNMP_COMMUNITY macro renamed to __SNMP_COMMUNITY

# Version 2.1 (2012-05-30)
    * All check_commands in templates directory have been given an okc- prefix
    * Deprecated host templates have been removed
    * default-host changed to okc-default-host
    * hostgroups removed

# Version 2 (2011-07-01):
    * All service names have been given an okc- prefix

# Version 1.1 (2011-07-01)
    * All "default service" created turned out to be registered

# Version 1 (2011-04-01):
    * templates now require a service called "$HOSTDESCR$", okc tools handle this automatically for new templates
    * Requires okconfig tools version 1
# Version 0:
    * Initial Release
"""


from __future__ import absolute_import
from __future__ import print_function
from pynag import Model
import okconfig

    
def get_template_version():
    """ Get the version number of the template directory """
    try:
        file = open("%s/metadata" % okconfig.template_directory)
        for line in file.readlines():
            line = line.split()
            if len(line) != 2: continue
            if line[0] == 'TEMPLATES_VERSION':
                return float(line[1])
        return 0
    except Exception:
        return 0



def upgrade_to_version_1_1():
    """ Hosts created with older addhost, might have some registered template services """
    print("Upgrading to config version 1.1 ...", end=' ')
    all_hosts = okconfig.get_hosts()
    my_services = Model.Service.objects.filter(register="1",name__contains='')
    for service in my_services:
        if service.name in all_hosts and not service['service_description']:
            service['register'] = 0
            print("Default service for %s updated" % service.name)
            service.save()
    print("ok")
def upgrade_to_version_2():
    """ Upgrades all nagios configuration file according to new okconfig templates
    
    Biggest change here is that all template services have been given a okc- prefix
    
    We have to find all services that have invalid parents.
    """
    print("Upgrading to config version 2.0 ...", end=' ')
    all_templates = Model.Service.objects.filter(name__contains='')
    all_templates = [x.name for x in all_templates]
    my_services = Model.Service.objects.filter(use__contains="")
    for service in my_services:
        old_use = service.use.split(',')
        new_use = []
        for i in old_use:
            okc_name = 'okc-%s' % i
            if not i in all_templates and okc_name in all_templates:
                # parent does not exist, but okc-parent does
                i = okc_name
            new_use.append( i )
        #new_use = ','.join(new_use)
        if old_use != new_use:
            service.use = ','.join(new_use)
            print(".. %s updated" % new_use)
            service.save()
    print("ok")
def upgrade_to_version_2_1():
    """ Upgrades all nagios configuration file according to new okconfig templates
    
    Biggest change here is that all template commands have been given a okc- prefix
    
    We have to find all services that have invalid check_command.
    """
    print("Upgrading to config version 2.1 ...", end=' ')
    all_commands = Model.Command.objects.all
    all_commands = [x.command_name for x in all_commands]
    my_services = Model.Service.objects.all
    for i in my_services:
        # Only work on services that actually define check_command
        if 'check_command' not in i._defined_attributes: continue
        
        # We only need the actual command, not any parameters
        check_command = i.check_command.split('!',1)[0]
        
        # If check_command is valid, then there is nothing to do
        if check_command in all_commands: continue
        
        # if command is missing, it might have been renamed by okconfig:
        if "okc-%s"%check_command in all_commands:
            i.check_command = "okc-%s"%i.check_command
            i.save()
            print("%s renamed to okc-%s"%(check_command,check_command))

    # Update notification commands on all contacts
    for i in Model.Contact.objects.all:
        current = i.service_notification_commands
        if current not in all_commands and "okc-%s"%current in all_commands:
            i.service_notification_commands = "okc-%s"%current
            i.save()
        current = i.host_notification_commands
        if current not in all_commands and "okc-%s"%current in all_commands:
            i.host_notification_commands = "okc-%s"%current
            i.save()
    # Check for hosts
    deprecated_hostnames = ('generic-server-dev', 'generic-server-prod', 'generic-server-crit', 'generic-server', 'default-host')
    new_hostname = "okc-default-host"
    
    all_templates = Model.Host.objects.filter(name__contains='')
    all_templates = [x.name for x in all_templates]
    my_hosts = Model.Host.objects.filter(use__contains="")
    for host in my_hosts:
        old_use = host.use.split(',')
        new_use = []
        for i in old_use:
            okc_name = 'okc-%s' % i
            if not i in all_templates and okc_name in all_templates:
                # parent does not exist, but okc-parent does
                i = okc_name
            elif not i in all_templates and i in deprecated_hostnames:
                i = new_hostname
            new_use.append( i )
        #new_use = ','.join(new_use)
        if old_use != new_use:
            host.use = ','.join(new_use)
            print(".. %s updated" % old_use)
            host.save() 
    print("ok")
def upgrade_to_version_2_2():
    """ Upgrades all nagios configuration file according to new okconfig templates

    Biggest Change here is that brocade templates previously had macro _SNMP_COMMUNITY which has been
    renamed to __SNMP_COMMUNITY
    """
    print("Upgrading to config version 2.2 ...", end=' ')
    my_services = Model.Service.objects.filter(use__contains='okc-brocade')
    for service in my_services:
        if service.host_name is None:
            continue
        if service['_SNMP_COMMUNITY'] in (None,'public'):
            continue
        if not service['__SNMP_COMMUNITY'] == 'public':
            continue
        print("..", service.get_description(), "renamed _SNMP_COMMUNITY %s to __SNMP_COMMUNITY" % (service['_SNMP_COMMUNITY']))
    print("ok")

def upgrade_to_version_2_3():
    """ Upgrade to version 2.3
     okc-emc_check_port_state now has a username and password field implied
    """
    print("Upgrading to config version 2.3 ...", end=' ')
    my_services = Model.Service.objects.filter(check_command="okc-emc-check_portstate", __USERNAME=None, register=1)
    for i in my_services:
        if i.get('__USERNAME') is None:
            i['__USERNAME'] = 'nagios'
        if i.get('__PASSWORD') is None:
            i['__PASSWORD'] = 'not set'
        i.save()
        print("Updated: ", i.get_shortname())

    # Add missing macros to okc-check_http
    my_services = Model.Service.objects.filter(check_command="okc-check_http", __VIRTUAL_HOST=None, register=1)
    for i in my_services:
        if i.get('__VIRTUAL_HOST') is None:
            i['__VIRTUAL_HOST'] = i.get('host_name')
        if i.get('__PORT') is None:
            i['__PORT'] = "80"
        i.save()
        print("Updated: ", i.get_shortname())

    my_services = Model.Service.objects.filter(check_command="okc-check_https", __VIRTUAL_HOST=None, register=1)
    for i in my_services:
        if i.get('__VIRTUAL_HOST') is None:
            i['__VIRTUAL_HOST'] = i.get('host_name')
        if i.get('__PORT') is None:
            i['__PORT'] = "443"
        i.save()
        print("Updated: ", i.get_shortname())

    my_services = Model.Service.objects.filter(check_command="okc-check_https_certificate", __PORT=None, register=1)
    for i in my_services:
        if i.get('__PORT') is None:
            i['__PORT'] = "443"
        i.save()
        print("Updated: ", i.get_shortname())
    print("ok")

def upgrade_to_version_2_4():
    """ Upgrade to version 2.4
    We will adapt to the following removals:
    * hostgroup cisco-hostgroup
    * servicegroup proliant-services
    * contactgroup proliant-contacts
    * servicegroup eva-services
    * contactgroup eva-contacts
    * servicegroup windows-services
    * servicegroup linux-services
    * servicegroup mssql-services
    """
    print("Upgrading to config version 2.4 ...", end=' ')
    dest_file = okconfig.config.destination_directory + "/backwards-compatibility.cfg"
    hostgroups = ['cisco-hostgroup']
    servicegroups = ['proliant-services', 'eva-services', 'windows-services', 'linux-services', 'mssql-services']
    contactgroups = ['proliant-contacts', 'eva-contacts']
    for i in hostgroups:
        hg = Model.Hostgroup.objects.filter(hostgroup_name=i)
        hg_hosts = Model.Host.objects.filter(host_groups__has_field=i)
        if hg_hosts and not hg:
            h = Model.Hostgroup()
            h['hostgroup_name'] = i
            h['alias'] = i
            h.set_filename(dest_file)
            h.save()
            print("Created hostgroup", i)
    for i in servicegroups:
        sg = Model.Servicegroup.objects.filter(servicegroup_name=i)
        sg_services = Model.Service.objects.filter(service_groups__has_field=i)
        if sg_services and not sg:
            s = Model.Servicegroup()
            s['servicegroup_name'] = i
            s['alias'] = i
            s.set_filename(dest_file)
            s.save()
            print("Created servicegroup", i)
    for i in contactgroups:
        cg = Model.Contactgroup.objects.filter(contactgroup_name=i)
        cg_contacts = Model.Contact.objects.filter(contactgroups__has_field=i)
        cg_hosts = Model.Host.objects.filter(contact_groups__has_field=i)
        cg_services = Model.Host.objects.filter(contact_groups__has_field=i)
        if not cg and (cg_contacts or cg_hosts or cg_services):
            c = Model.Contactgroup()
            c['contactgroup_name'] = i
            c['alias'] = i
            c.set_filename(dest_file)
            c.save()
            print("Created contactgroup", i)
    print("ok")

def upgrade_to_version_2_5():
    """ Upgrade to version 2.5

    We will adapt to the following changes:
    * okc-check_http* has a new service variable __ON_REDIRECT

    """
    print("Upgrading to config version 2.5 ...", end=' ')
    services = Model.Service.objects.filter(check_command__startswith='okc-check_http', __ON_REDIRECT__exists=False)
    for i in services:
        i['__ON_REDIRECT'] = "follow"
        i.save()
    print("ok")




def rename_oktemplate_services():
    """ To change config version to 2.0 This is a one-off action. Not part of any upgrade """
    all_obj = Model.Service.objects.filter(name__contains='',filename__startswith=okconfig.template_directory)
    for i in all_obj:
        if i.name.startswith('okc-'): continue
        i['name'] = "okc-%s" % i.name
        i.save()    



def upgrade_okconfig():
    """Upgrades nagios configuration to match the level of current oktemplates format"""
    template_version = get_template_version()
    print("Upgrading to version %s" % template_version)
    if template_version >= 1:
        # "We dont need to do anything, the tools have been upgraded"
        pass
    if template_version >= 1.1:
        upgrade_to_version_1_1()
    if template_version >= 2:
        upgrade_to_version_2()
    if template_version >= 2.1:
        upgrade_to_version_2_1()
    if template_version >= 2.2:
        upgrade_to_version_2_2()
    if template_version >= 2.3:
        upgrade_to_version_2_3()
    if template_version >= 2.4:
        upgrade_to_version_2_4()
    if template_version >= 2.5:
        upgrade_to_version_2_5()

if __name__ == '__main__':
    upgrade_okconfig()



#
# Copyright 2011, Pall Sigurdsson <palli@opensource.is>
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from builtins import object
import subprocess
import socket
from pynag import Model
from .helper_functions import runCommand
import okconfig

class ScannedHost(object):
    """Simple datastructure for a recently portscanned host"""
    def __init__(self, ipaddress=None, hostname=None,ismonitored=None):
        self.ipaddress = ipaddress
        self.hostname = hostname
        self.ismonitored = ismonitored
        if hostname is None:
            self.hostname = self.get_hostname()
    def check(self):
        """Runs all self.ch*. Returns nothing"""

        if self.is_windows():
            self.platform = "Windows"
        elif self.is_linux():
            self.platform = "Linux"
        else:
            self.platform = "Unknown"

        if self.is_agent_okconfig():
            self.nrpe = "OKconfig agent"
        elif self.is_agent_responding():
            self.nrpe = "Generic NRPE running"
        elif self.is_agent_installed():
            self.nrpe = "NRPE running but does not respond to us"
        else:
            self.nrpe = "NRPE Not detected"

        if self.has_webserver():
            self.port80 = "yes"
        else:
            self.port80 = "no"

    def get_hostname(self):
        """returns hostname given a specific ip address"""
        try:
            name= socket.gethostbyaddr(self.ipaddress)
            return name[0]
        except Exception:
            return None
    def is_windows(self):
        """Returns true if machine replies on port 3389"""
        return check_tcp(host=self.ipaddress, port=3389)
    def is_agent_installed(self):
        """Returns True if nrpe client is running"""
        return check_tcp(host=self.ipaddress, port=5666)
    def is_linux(self):
        """Returns true if machine replies on port 22"""
        return check_tcp(host=self.ipaddress, port=22)
    def has_webserver(self):
        """Returns true if machine replies on port 80"""
        return check_tcp(host=self.ipaddress, port=80)
    def has_mssql(self):
        """Returns true if machine replies on port 1433"""
        return check_tcp(host=self.ipaddress, port=1433)
    def has_nrpe(self):
        """Returns true if machine replies on port 5666"""
        return check_tcp(host=self.ipaddress, port=5666)
    def has_ssl(self):
        """Returns true if machine replies on port 443"""
        return check_tcp(host=self.ipaddress, port=443)

    def is_agent_responding(self):
        """returns true if host responds to check_nrpe commands"""
        returncode, stdout, stderr = runCommand("check_nrpe -H '%s'" % self.ipaddress)
        if returncode == 0:
            return True
        if returncode == 1:
            return False
        return None
    def is_agent_okconfig(self):
        """returns true if host responds to check_nrpe commands"""
        returncode,stdout,stderr = runCommand("check_nrpe -H '%s' -c get_disks")
        if returncode == 0: return True
        if returncode == 1: return False
        return None
def check_tcp(host="localhost", port=22, timeout=5):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(timeout)
        s.connect( (host, port) )
        #s.send('')
        #s.recv(1024)
        s.close()
        return True
    except Exception:
        #raise e
        return False

def runCommand(command):
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE,)
    stdout, stderr = proc.communicate('through stdin to stdout')
    returncode = proc.returncode
    return returncode,stdout,stderr


def get_ip_address_list():
    """returns a list of every ip address of every host in nagios"""
    hosts = Model.Host.objects.all
    result = []
    for host in hosts:
        a = host['address']
        if not a is None and a not in result:
            result.append(a)
    return result

def check_nrpe(host):
    command = "check_nrpe -H '%s'" % host

def get_my_ip_address():
    """Returns default ip address of this host"""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('94.142.154.11', 80))
    return s.getsockname()[0]

def pingscan(network='192.168.1.0/24'):
    """scans a specific network, returns a list of all ip that respond"""
    command =  "fping -t 50 -i 10 -a "
    if network.find('/') > 0: command += " -g "
    command += network
    r,stdout,stderr = runCommand(command)
    if r > 1:
        raise Exception("Error running %s: %s" % (command,stderr) )
    ip_list = []
    for i in stdout.split('\n'):
        try: socket.inet_aton(i)
        except Exception: continue
        ip_list.append(i)
    return ip_list

def get_new_hosts(network='192.168.1.0/24'):
    """Returns a list of ScannedHost alive in given network that do not exist in nagios"""
    registered_ips = get_ip_address_list()
    new_hosts =[]
    for i in get_all_hosts(network=network):
        if not i.ipaddress in registered_ips:
            new_hosts.append(i)
    return new_hosts

def get_all_hosts(network='192.168.1.0/24'):
    """Returns a list of ScannedHost alive in given network"""
    scanned_ips = pingscan(network)
    registered_ips = get_ip_address_list()
    all_hosts = []
    for i in scanned_ips:
            ismonitored = i in registered_ips
            host = ScannedHost(ipaddress=i,ismonitored=ismonitored)
            all_hosts.append( host)
    return all_hosts

def get_hostname(self):
    """returns hostname given a specific ip address"""
    try:
        name= socket.gethostbyaddr(self.ipaddress)
        return name[0]
    except Exception:
        return None

def traceroute(host="localhost"):
    """ Returns a list of every host discovered while tracerouting

    Returns:
        ["ip1","ip2","ip3",...]
    """
    result = []
    returncode,stdout,stderr = runCommand("traceroute -n '%s'" % host)
    print(stdout)
    if returncode > 0:
        raise okconfig.OKConfigError("Failed to traceroute host %s, error: %s" % (host, stderr))
    for line in stdout.split("\n"):
        line = line.split()
        if len(line) < 2:
            continue
        if line[0].isdigit():
            result.append(line[1])
    return result

def set_network_parents(host_name, method="traceroute"):
    """ Autocreates network parents for a given host """
    if method == 'traceroute':
        parents = traceroute(host=host_name)
        print(len(parents), "parents found")
        previous_host = None
        for i in parents:
            hosts = Model.Host.objects.filter(address=i)
            if len(hosts) == 0:
                print("adding host ", i)
                dnsname = get_hostname(i)
                if dnsname is None:
                    dnsname = i
                okconfig.addhost(host_name=dnsname, address=i)
                hosts = Model.Host.objects.filter(address=i)
            host = hosts[0]
            if previous_host is not None:
                print(previous_host.address, "->", host.address)
                host.parents = previous_host.host_name
                host.save()
            previous_host = host


if __name__ == '__main__':
    print(set_network_parents('mbl.is'))


# setup.py ###
from __future__ import absolute_import
from __future__ import print_function
from setuptools import setup
import os

NAME = "okconfig"
VERSION = '1.4.2'
SHORT_DESC = "%s - Powertools to generate Nagios configuration files" % NAME
LONG_DESC = """
%s contains tools and templates for enterprise quality nagios configuration.
""" % NAME


def get_filelist(path):
    """Returns a list of all files in a given directory"""
    files = []
    directories_to_check = [path]
    while len(directories_to_check) > 0:
        current_directory = directories_to_check.pop(0)
        for i in os.listdir(current_directory):
            if i == '.gitignore':
                continue
            relative_path = current_directory + "/" + i
            if os.path.isfile(relative_path):
                files.append(relative_path)
            elif os.path.isdir(relative_path):
                directories_to_check.append(relative_path)
            else:
                print("what am i?", i)
    return files

if __name__ == "__main__":
    manpath = "share/man/man1/"
    etcpath = "/etc/%s" % NAME
    etcmodpath = "/etc/%s/modules" % NAME
    initpath = "/etc/init.d/"
    logpath = "/var/log/%s/" % NAME
    varpath = "/var/lib/%s/" % NAME
    rotpath = "/etc/logrotate.d"
    datarootdir = "/usr/share/%s" % NAME
    template_files = get_filelist('usr/share/okconfig')
    data_files = [("/" + os.path.dirname(x), [x]) for x in template_files]
    data_files.append((logpath, []))
    data_files.append((rotpath, ["etc/logrotate.d/okconfig"]))
    #data_files.append(( manpath, ['docs/okconfig.1.gz'] ))
    setup(
        name='%s' % NAME,
        version=VERSION,
        author='Pall Sigurdsson',
        description=SHORT_DESC,
        long_description=LONG_DESC,
        author_email='palli@opensource.is',
        maintainer='Gardar Thorsteinsson',
        maintainer_email='gardar@ok.is',
        url='http://okconfig.org',
        license='GPL',
        scripts=['usr/bin/okconfig'],
        packages=['okconfig'],
        requires=['unittest2'],
        install_requires=['paramiko','pynag','six','future','setuptools_rust'],
        data_files=data_files,
    )



#
# Copyright 2011, Pall Sigurdsson <palli@opensource.is>
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module provides access to default configuration values in /etc/okconfig.conf """

from __future__ import absolute_import
from builtins import str
import os

# This is the main configuration file
config_file             = "/etc/okconfig.conf"

# First some default values, in case config does not specify any
nagios_config           = "/etc/nagios/nagios.cfg"
template_directory      = "/etc/nagios/okconfig/templates"
examples_directory      = "/etc/nagios/okconfig/examples"
destination_directory   = "/etc/nagios/okconfig/"
install_nrpe_script     = "/usr/share/okconfig/client/linux/install_okagent.sh"
nsclient_installfiles   = "/usr/share/okconfig/client/windows/"
examples_directory_local= destination_directory + "/examples"
git_commit_changes      = "1"
try:
    if os.path.isfile(config_file):
        for line in open(config_file).readlines():
            line = line.strip()
            if line.startswith('#'): continue
            line = line.split(None,1)
            if len(line) != 2: continue
            keyword = str(line[0]).strip()
            value = str(line[1]).strip()
            if keyword   == "nagios_config": nagios_config = value
            elif keyword == "template_directory": template_directory = value
            elif keyword == "examples_directory": examples_directory = value
            elif keyword == "destination_directory": destination_directory = value
            elif keyword == "install_nrpe_script": install_nrpe_script = value
            elif keyword == "nsclient_installfiles": nsclient_installfiles = value
            elif keyword == "examples_directory_local": examples_directory_local = value
            elif keyword == "git_commit_changes": git_commit_changes = value
except ImportError:
    raise
